# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Saturation Environment
=======================
"""  # noqa: D205, D400

import random
from abc import abstractmethod
from typing import Any, Optional

from gymnasium import Env, spaces
from gymnasium.spaces.text import alphanumeric

from gym_saturation.constants import FALSEHOOD_SYMBOL, MOCK_TPTP_PROBLEM

ALPHANUMERIC_WITH_UNDERSCORE = "".join(alphanumeric) + "_"
EXTENDED_ALPHANUMERIC = ALPHANUMERIC_WITH_UNDERSCORE + "(), |~=!$.'"


class SaturationEnv(Env[tuple[str, ...], str]):
    """
    Saturation algorithm in a reinforcement learning friendly way.

    It's an abstract class, so here we have only trivial smoke tests.
    One should override ``_do_deductions`` method in children classes.

    Refer to :ref:`saturation_env` for more documentation.

    >>> class DummyProver(SaturationEnv):
    ...     def _do_deductions(action):
    ...         pass

    >>> env = DummyProver()
    """

    reward_range = (0, 1)
    action_space = spaces.Text(256, charset=ALPHANUMERIC_WITH_UNDERSCORE)
    observation_space = spaces.Sequence(
        spaces.Text(4000, charset=EXTENDED_ALPHANUMERIC)
    )

    def __init__(  # noqa: D107
        self,
    ):
        super().__init__()
        self._task = MOCK_TPTP_PROBLEM
        self._terminated = False
        self._available_actions: set[str] = set()

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict[str, Any]] = None,
    ) -> tuple[tuple[str, ...], dict[str, Any]]:
        """
        Reset the environment.

        :param seed: seed for compatibility
        :param options: options for compatibility
        :returns: observations and info
        """
        super().reset(seed=seed)
        random.seed(seed)
        self._terminated = False
        self._available_actions = set()
        return (), {}

    @abstractmethod
    def _do_deductions(self, action: Any) -> tuple[tuple[str, ...], set[str]]:
        raise NotImplementedError  # pragma: no cover

    def step(
        self, action: Any
    ) -> tuple[tuple[str, ...], float, bool, bool, dict[str, Any]]:
        """
        Run one time-step of the environment's dynamics.

        When end of episode is reached, you are responsible for calling
        ``reset()`` to reset this environment's state.
        Accepts an action and returns a tuple
        (observation, reward, terminated, truncated, info)

        :param action: an action provided by the agent
        :returns: a tuple of four values:\n
            * observation: agent's observation of the current environment
            * reward: amount of reward returned after previous action
            * terminated: whether the proof was found
            * truncated: whether the episode was finished for an external
              reason (e.g. time limit)
            * info: contains auxiliary diagnostic information (helpful for
              debugging, and sometimes learning)
        """  # noqa: D301
        new_clauses: tuple[str, ...] = ()
        if not self._terminated and action in self._available_actions:
            new_clauses, new_actions = self._do_deductions(action)
            self._terminated = max(
                (FALSEHOOD_SYMBOL in clause for clause in new_clauses),
                default=False,
            )
            self._available_actions.discard(action)
            self._available_actions.update(new_actions)
        return (
            new_clauses,
            0.0,
            self._terminated,
            False,
            {},
        )

    def render(self) -> None:
        """No render."""

    def set_task(self, task: str) -> None:
        """
        Set the specified task to the current environment.

        :param task: a TPTP problem filename
        """
        self._task = task

    def get_task(self) -> str:
        """
        Get the task that the agent is performing in the current environment.

        :returns: a TPTP problem filename
        """
        return self._task



# Copyright 2023-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Constants used throughout the package
======================================
"""  # noqa: D205, D400

import os
from importlib.resources import files

CLAUSE_EMBEDDINGS = "clause_embeddings"
FALSEHOOD_SYMBOL = "$false"
MOCK_TPTP_FOLDER = str(
    files("gym_saturation").joinpath(os.path.join("resources", "TPTP-mock"))
)
MOCK_TPTP_PROBLEM = os.path.join(
    MOCK_TPTP_FOLDER, "Problems", "TST", "TST001-1.p"
)



# Copyright 2022-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Saturation Environment with iProver back-end
============================================
"""  # noqa: D205, D400

import asyncio
import json
import os
import re
from threading import Thread
from typing import Any, Optional

from gym_saturation.envs.saturation_env import SaturationEnv
from gym_saturation.relay_server import (
    QUERY_END_MESSAGE,
    SESSION_END_MESSAGE,
    RelayServer,
    RelayTCPHandler,
)


def _iprover_start(
    iprover_port: int, problem_filename: str, prover_binary_path: str
) -> asyncio.subprocess.Process:
    tptp_folder = os.path.join(os.path.dirname(problem_filename), "..", "..")
    command = " ".join(
        [
            prover_binary_path,
            "--interactive_mode",
            "true",
            "--external_ip_address",
            "127.0.0.1",
            "--external_port",
            str(iprover_port),
            "--schedule",
            "none",
            "--resolution_flag",
            "false",
            "--instantiation_flag",
            "false",
            "--superposition_flag",
            "true",
            "--sup_iter_deepening",
            "0",
            "--sup_passive_queue_type",
            "external_agent",
            "--preprocessing_flag",
            "false",
            "--include_path",
            tptp_folder,
            problem_filename,
        ]
    )
    return asyncio.run(
        asyncio.create_subprocess_shell(
            command,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
    )


class IProverEnv(SaturationEnv):
    """
    An RL environment around iProver.

    :param prover_binary_path: a path to iProver binary;
        by default, we assume it to be ``iproveropt`` and in the $PATH

    Refer to :ref:`saturation_env` for more documentation.

    >>> from gymnasium.utils.env_checker import check_env
    >>> import gymnasium as gym
    >>> env = gym.make(
    ...     "iProver-v0",
    ... ).unwrapped
    >>> env.relay_server
    Traceback (most recent call last):
     ...
    ValueError: run ``reset`` first!
    >>> check_env(env)

    episode is never truncated by default

    >>> _ = env.reset()
    >>> _, _, _, truncated, _ = env.step("c_49")
    >>> truncated
    False
    """

    def __init__(  # noqa: D107
        self,
        prover_binary_path: str = "iproveropt",
    ):
        super().__init__()
        self.prover_binary_path = prover_binary_path
        self._relay_server: Optional[RelayServer] = None
        self.relay_server_thread: Optional[Thread] = None
        self.iprover_process: Optional[asyncio.subprocess.Process] = None

    def _restart_relay_server(self) -> None:
        if self._relay_server:
            self._terminate_threads()
        self._relay_server = RelayServer(("localhost", 0), RelayTCPHandler)
        self.relay_server_thread = Thread(
            target=self._relay_server.serve_forever
        )
        self.relay_server_thread.daemon = True
        self.relay_server_thread.start()

    def _parse_batch_clauses(
        self, batch_clauses: list[dict[str, Any]]
    ) -> tuple[tuple[str, ...], set[str]]:
        new_labels: set[str] = set()
        new_clauses: tuple[str, ...] = ()
        for dict_clause in batch_clauses:
            raw_clause = (
                dict_clause["clause"].replace("\n", "").replace(" ", "")
            )
            (label, literals, inference_record) = re.findall(
                pattern=r"cnf\((\w+),\w+,\((.*)\),(\w+\(.+\))\)\.",
                string=raw_clause,
            )[0]
            new_labels.add(label)
            if inference_record[:5] == "file(":
                new_clauses += (
                    f"cnf({label},axiom,{literals},file('input.p')).",
                )
            else:
                new_clauses += (
                    f"cnf({label},plain,{literals},"
                    f"{inference_record.replace('status(thm)', '')}).",
                )
        return new_clauses, new_labels

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict[str, Any]] = None,
    ) -> tuple[tuple[str, ...], dict[str, Any]]:
        """
        Reset the environment.

        :param seed: seed for compatibility
        :param options: options for compatibility
        :returns: observations and info
        """
        super().reset(seed=seed)
        self._restart_relay_server()
        self.iprover_process = _iprover_start(
            self.relay_server.server_address[1],
            self.get_task(),
            self.prover_binary_path,
        )
        data = self._get_json_data()
        new_clauses, new_labels = self._parse_iprover_requests(data)
        self._available_actions = new_labels
        return new_clauses, {}

    def _get_json_data(self) -> list[dict[str, Any]]:
        json_data = [{"tag": "None"}]
        while json_data[-1]["tag"] not in {
            QUERY_END_MESSAGE,
            SESSION_END_MESSAGE,
        }:
            parsed_json = self.relay_server.input_queue.get()
            self.relay_server.input_queue.task_done()
            json_data.append(parsed_json)
        return json_data[1:]

    def _parse_iprover_requests(
        self, iprover_requests: list[dict[str, Any]]
    ) -> tuple[tuple[str, ...], set[str]]:
        new_labels: set[str] = set()
        new_clauses: tuple[str, ...] = ()
        for iprover_request in iprover_requests:
            if "clauses" in iprover_request:
                more_new_clauses, more_new_labels = self._parse_batch_clauses(
                    iprover_request["clauses"]
                )
                new_labels.update(more_new_labels)
                new_clauses += more_new_clauses
        return new_clauses, new_labels

    def _do_deductions(self, action: str) -> tuple[tuple[str, ...], set[str]]:
        iprover_scores_message = {
            "tag": "given_clause_res",
            "passive_is_empty": False,
            "given_clause": int(action[2:]),
        }
        relayed_scores_message = bytes(
            json.dumps({"tag": "server_queries_end"})
            + "\n\x00\n"
            + json.dumps(iprover_scores_message)
            + "\n\x00\n",
            "utf8",
        )
        self.relay_server.output_queue.put(relayed_scores_message)
        data = self._get_json_data()
        return self._parse_iprover_requests(data)

    def close(self) -> None:
        """Stop relay server."""
        self._terminate_threads()

    @property
    def relay_server(self) -> RelayServer:
        """
        Return the relay server object.

        :raises ValueError: if called before ``reset``
        """
        if self._relay_server:
            return self._relay_server
        raise ValueError("run ``reset`` first!")

    def _terminate_threads(self) -> None:
        if self._relay_server:
            self._relay_server.shutdown()
        if self.relay_server_thread:
            self.relay_server_thread.join()
        if self.iprover_process:
            try:
                self.iprover_process.terminate()
            except ProcessLookupError:  # pragma: no cover
                pass



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Vampire Wrapper
================
"""  # noqa: D205, D400

from typing import Optional

import pexpect


class VampireWrapper:
    """
    A wrapper around Vampire binary running in a manual clause selection mode.

    :param binary_path: an absolute path to Vampire binary
    :param command_line_arguments: command line arguments in one string

    .. _vampire-wrapper :

    >>> from gym_saturation.constants import (MOCK_TPTP_PROBLEM,
    ...     MOCK_TPTP_FOLDER)
    >>> vampire = VampireWrapper("vampire")
    >>> vampire.pick_a_clause("2")
    Traceback (most recent call last):
     ...
    ValueError: start solving a problem first!
    >>> result = vampire.start(MOCK_TPTP_PROBLEM, MOCK_TPTP_FOLDER)
    """

    def __init__(  # noqa: D107
        self, binary_path: str, command_line_arguments: Optional[str] = None
    ):
        self.binary_path = binary_path
        self._proc = None
        self.problem_filename: Optional[str] = None
        self.command_line_arguments = (
            " --manual_cs on --show_passive on"
            " --show_new on --time_limit 0 --avatar off "
            if command_line_arguments is None
            else command_line_arguments
        )

    def _get_stdout(self) -> tuple[tuple[str, str, str], ...]:
        result: tuple[tuple[str, str, str], ...] = ()
        self.proc.expect(
            ["Pick a clause:", "Pick a clause pair:", pexpect.EOF]
        )
        if self.proc.before is not None:
            lines = self.proc.before.decode("utf-8").split("\r\n")
            for line in lines:
                if line[:5] == "[SA] ":
                    result_type, result_body = line[5:].split(": ")
                    clause_label, clause = result_body.split(". ")
                    result += ((result_type, clause_label, clause),)
        return result

    def start(
        self, problem_filename: str, tptp_folder: str
    ) -> tuple[tuple[str, str, str], ...]:
        """
        Start Vampire in a manual mode on a given problem.

        Time limit is one day, Vampire prints everything

        :param problem_filename: full path of a TPTP problem file
        :param tptp_folder: the root folder for TPTP library
        :returns: a sequence of action type, clause number and clause
        """
        self.problem_filename = problem_filename
        if self._proc is not None:
            self._proc.close()
        self._proc = pexpect.spawn(
            f"{self.binary_path} {self.command_line_arguments} "
            f"--include {tptp_folder} {problem_filename}",
            echo=False,
        )
        # https://pexpect.readthedocs.io/en/stable/commonissues.html#timing-issue-with-send-and-sendline
        self._proc.delaybeforesend = None  # type: ignore
        return self._get_stdout()

    def pick_a_clause(
        self, clause_label: str
    ) -> tuple[tuple[str, str, str], ...]:
        """
        Select a clause and get response from Vampire.

        :param clause_label: a given clause order number
        :returns: a sequence of action type, clause number and clause
        """
        self.proc.sendline(clause_label)
        return self._get_stdout()

    @property
    def proc(self) -> pexpect.spawn:
        """
        Vampire process.

        :raises ValueError: when called before ``reset``
        """
        if self._proc is None:
            raise ValueError("start solving a problem first!")
        return self._proc

    def terminate(self) -> None:
        """Terminate Vampire process if any."""
        if self._proc is not None:
            self._proc.terminate()
            self._proc.wait()



"""Labels extractor wrapper."""

import re

import gymnasium as gym


class LabelsExtractor(gym.ObservationWrapper):
    """
    Labels extractor wrapper.

    >>> from gym_saturation.envs.vampire_env import VampireEnv
    >>> env = LabelsExtractor(VampireEnv())
    >>> observation, info = env.reset()
    >>> type(observation)
    <class 'dict'>
    >>> observation.keys()
    dict_keys(['labels', 'observation'])
    >>> type(observation["labels"])
    <class 'tuple'>
    >>> type(observation["labels"][0])
    <class 'str'>
    """

    def observation(
        self, observation: tuple[str, ...]
    ) -> dict[str, tuple[str, ...]]:
        """
        Return a modified observation.

        :param observation: The observation
        :returns: The modified observation
        """
        return {
            "labels": tuple(
                re.findall(r"cnf\((\w+),.+\)\.", clause)[0]
                for clause in observation
            ),
            "observation": observation,
        }



# Copyright 2022-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Relay Server between Two Sockets
================================
"""  # noqa: D205, D400

import json
from queue import Queue
from socketserver import BaseRequestHandler, ThreadingTCPServer
from typing import Any

QUERY_END_MESSAGE = "server_queries_start"
SESSION_END_MESSAGE = "proof_out"


class RelayServer(ThreadingTCPServer):
    r"""
    Server relaying i/o of a long-running connection to a TCP socket to queues.

    >>> import socket
    >>> from threading import Thread
    >>> with RelayServer(("localhost", 0), RelayTCPHandler
    ... ) as relay_server:
    ...     thread = Thread(target=relay_server.serve_forever)
    ...     thread.daemon = True
    ...     thread.start()
    ...     with socket.create_connection(relay_server.server_address
    ...     ) as socket_connection:
    ...         # data sent to the socket are stored in one queue
    ...         socket_connection.sendall(bytes(
    ...             f'{{"tag": "{QUERY_END_MESSAGE}"}}\n\x00\n', "utf8")
    ...         )
    ...         print(relay_server.input_queue.get())
    ...         # data from another queue are sent to the client
    ...         relay_server.output_queue.put(b"test")
    ...         print(str(socket_connection.recv(4096), "utf8"))
    ...         # message format for closing connection
    ...         socket_connection.sendall(bytes(
    ...             f'{{"tag": "{SESSION_END_MESSAGE}"}}\n\x00\n', "utf8"
    ...         ))
    ...     relay_server.shutdown()
    ...     thread.join()
    {'tag': 'server_queries_start'}
    test
    """

    def __init__(  # noqa: D107
        self,
        server_address: tuple[str, int],
        request_handler_class: type[BaseRequestHandler],
        bind_and_activate: bool = True,
    ):
        super().__init__(
            server_address, request_handler_class, bind_and_activate
        )
        self.input_queue: Queue = Queue()
        self.output_queue: Queue = Queue()
        self.daemon_threads = True


class RelayTCPHandler(BaseRequestHandler):
    """The request handler class for relay server."""

    def read_messages(self) -> list[dict[str, Any]]:
        """
        Read messages from TCP request.

        :returns: parsed messages
        """
        raw_data = ""
        json_messages: list[dict[str, Any]] = []
        while len(json_messages) == 0 or json_messages[-1]["tag"] not in {
            QUERY_END_MESSAGE,
            SESSION_END_MESSAGE,
        }:
            raw_data += str(self.request.recv(4096), "utf8")
            raw_jsons = raw_data.split("\n\x00\n")
            raw_data = raw_jsons[-1]
            json_messages.extend(list(map(json.loads, raw_jsons[:-1])))
        return json_messages

    def handle(self) -> None:
        """Read data from another TCP socket or send data to it."""
        if isinstance(self.server, RelayServer):
            json_messages: list[dict[str, Any]] = []
            while (
                len(json_messages) == 0
                or json_messages[-1]["tag"] != SESSION_END_MESSAGE
            ):
                json_messages = self.read_messages()
                for json_message in json_messages:
                    self.server.input_queue.put(json_message)
                if json_messages[-1]["tag"] == QUERY_END_MESSAGE:
                    self.request.sendall(self.server.output_queue.get())
                    self.server.output_queue.task_done()



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Saturation Environment with Vampire back-end
============================================
"""  # noqa: D205, D400

import os
import re
from typing import Any, Optional

from gym_saturation.constants import FALSEHOOD_SYMBOL
from gym_saturation.envs.saturation_env import SaturationEnv
from gym_saturation.vampire_wrapper import VampireWrapper


class VampireEnv(SaturationEnv):
    """
    An RL environment wrapper around Vampire prover.

    :param prover_binary_path: a path to Vampire binary;
        by default we expect it to be in the $PATH

    Refer to :ref:`saturation_env` for more documentation.

    We can run a full Gymnasium environment check:

    >>> from gymnasium.utils.env_checker import check_env
    >>> import gymnasium as gym
    >>> env = gym.make("Vampire-v0").unwrapped
    >>> check_env(env)

    repeating actions change nothing

    >>> env = gym.make("Vampire-v0")
    >>> _ = env.reset()
    >>> one = env.step("c_1")
    >>> two = env.step("c_1")
    >>> one == two
    True

    sometimes Vampire can solve a problem during pre-processing

    >>> from gym_saturation.constants import MOCK_TPTP_PROBLEM
    >>> trivial_problem = os.path.join(os.path.dirname(MOCK_TPTP_PROBLEM),
    ...     "TST002-1.p")
    >>> env.unwrapped.set_task(trivial_problem)
    >>> _, _ = env.reset()
    >>> env.unwrapped._terminated
    True
    >>> _, _, terminated, _, _ = env.step("anything")
    >>> terminated
    True
    >>> env.close()

    a test of an unexpected reply from Vampire

    >>> from gym_saturation.constants import MOCK_TPTP_FOLDER
    >>> vampire_binary = os.path.join(MOCK_TPTP_FOLDER, "..", "vampire-mock")
    >>> vampire_env = VampireEnv(prover_binary_path=vampire_binary)
    >>> vampire_env.reset()
    Traceback (most recent call last):
     ...
    ValueError: ('Unexpected response type: ', 'who could expect that?')
    """

    def __init__(  # noqa: D107
        self,
        prover_binary_path: str = "vampire",
    ):
        super().__init__()
        self._vampire = VampireWrapper(prover_binary_path)

    def _parse_vampire_response(
        self, vampire_response: tuple[tuple[str, str, str], ...]
    ) -> tuple[tuple[str, ...], set[str]]:
        new_labels: set[str] = set()
        new_clauses: tuple[str, ...] = ()
        for response_type, clause_label, clause_text in vampire_response:
            new_label = "c_" + clause_label
            if response_type == "passive" or FALSEHOOD_SYMBOL in clause_text:
                new_clauses += (
                    self._parse_vampire_clause(new_label, clause_text),
                )
                new_labels.add(new_label)
            elif response_type not in {
                "active",
                "forward reduce",
                "backward reduce",
                "new propositional",
                "new",
                "final",
                "input",
                "fn def discovered",
            }:
                raise ValueError("Unexpected response type: ", response_type)
        return new_clauses, new_labels

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict[str, Any]] = None,
    ) -> tuple[tuple[str, ...], dict[str, Any]]:
        """
        Reset the environment.

        :param seed: seed for compatibility
        :param options: options for compatibility
        :returns: observations and info
        """
        super().reset(seed=seed)
        tptp_folder = os.path.join(
            os.path.dirname(self.get_task()), "..", ".."
        )
        vampire_response = self._vampire.start(self.get_task(), tptp_folder)
        new_clauses, new_labels = self._parse_vampire_response(
            vampire_response
        )
        self._terminated = max(
            (FALSEHOOD_SYMBOL in clause for clause in new_clauses),
            default=False,
        )
        self._available_actions = new_labels
        return new_clauses, {}

    def _do_deductions(self, action: str) -> tuple[tuple[str, ...], set[str]]:
        return self._parse_vampire_response(
            # the first two characters are `c_`
            self._vampire.pick_a_clause(action[2:])
        )

    def _parse_vampire_clause(
        self, clause_label: str, clause_text: str
    ) -> str:
        literals, inference_rule, inference_parents = re.findall(
            r"(.+) \[([^\d,]+)([\d,]*)\]", clause_text
        )[0]
        literals = literals.replace(" ", "")
        inference_rule = inference_rule.strip().replace(" ", "_")
        inference_parents = "c_" + inference_parents.replace(",", ",c_")
        if inference_rule != "input":
            return (
                f"cnf({clause_label},plain,{literals},"
                f"inference({inference_rule},"
                f"[],[{inference_parents}]))."
            )
        return f"cnf({clause_label},axiom,{literals},file('input.p'))."

    def close(self) -> None:
        """Terminate Vampire process."""
        self._vampire.terminate()



# Copyright 2021-2025 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
There are two environments in this module.

They are registered using a limit for the number of steps in an episode and the
maximal possible reward is set to ``1.0`` (proof is found).
"""

from gymnasium.envs.registration import register

register(id="Vampire-v0", entry_point="gym_saturation.envs:VampireEnv")
register(id="iProver-v0", entry_point="gym_saturation.envs:IProverEnv")
